###########################
## R code for:
## (1) a function for simulating pest population dynamics assuming different gene-drive strategies and demographies
## (2) running the function for different parameter inputs and plotting the results
###########################

rm(list=ls())
library(ggplot2)
library(reshape2)
library(compiler)
library(iterators)
library(snow)
library(doSNOW)
library(foreach)

###########################################################
## Set up parallel processing
## nproc is the number of processing cores to use
###########################################################

nproc <- 6
cl.tmp = makeCluster(rep('localhost',nproc), type='SOCK')
registerDoSNOW(cl.tmp)
getDoParWorkers()

######################################
### Gene-drive simulation function
######################################

gd <- function(input) {
  
  ######################################
  ## assign all variables in input
  ######################################
  
  for (i in 1:ncol(input)) {assign(names(input)[i], input[,i])}
  
  ######################################
  ## Functions
  ######################################
  
  ## calculate population growth rate from population projection matrix
  r.func <- function(x) log(Re((eigen(x)$values)[1]))
  
  ## calculate stable age distibution from population projection matrix
  stable.age.dist <- function(x) ((x %*% (Re((eigen(x)$vectors)[,1])))/(sum((x %*% (Re((eigen(x)$vectors)[,1]))))))[,1]
  
  ## Classify phenotypic sex and reproductive status
  reproductive <- function(pop, strategy) {
    
    if (nrow(pop)>0) {
      
      if (strategy==1) {
        pop$p.sex <- ifelse(pop$g.sex=='f' & (pop$aut.m==-999 | pop$aut.p==-999),'m',pop$g.sex) ## sex reversal for females, only 1 gene-drive copy needed
        pop$reprod <- ifelse(pop$g.sex=='f' & (pop$aut.m==-999 | pop$aut.p==-999),0,1) ## only females are made sterile by gene drive
      } else if (strategy==2) {
        pop$p.sex <- ifelse(pop$g.sex=='f' & (pop$aut.m==-999 | pop$aut.p==-999),'m',pop$g.sex) ## sex reversal for females, only 1 gene-drive copy needed
        pop$reprod <- 1 ## no sterility
      } else if (strategy==3) {
        pop$p.sex <- pop$g.sex
        pop$reprod <- 1
      } else if (strategy==4) {
        pop$p.sex <- pop$g.sex
        if (position=='intron') {
          pop$reprod <- ifelse(pop$g.sex=='f' & pop$aut.m==-999 & pop$aut.p==-999,0,1) 
        } else {
          pop$reprod <- ifelse(pop$g.sex=='f' & pop$aut.m.func==0 & pop$aut.p.func==0,0,1) 
        }
      }
      
    }
    
    pop
  }
  
  ## Create population
  create <- function(nInit, nInit.vec, nG, Pr){
    
    pop <- data.frame(id=1:nInit)
    pop$age <- rep(rep(c(genTime,genTime*2),2),nInit.vec)
    
    ## genotypic and phenotypic sex
    pop$g.sex <- c('f','m')[rep(1:2,c(sum(nInit.vec[1:2]),sum(nInit.vec[3:4])))] ## genotypic sex
    pop$p.sex <- pop$g.sex ## phenotypic sex
    
    ## Autosomes
    pop$aut.p <- pop$aut.m <- nGuides ## start with full complement of cutting sites
    
    ## Allosomes
    pop$all.m <- 'X' ## maternal allosome (sex chromosome)
    pop$all.p <- ifelse(pop$g.sex=='f','X','Y') ## paternal allosome (sex chromosome)
    
    ## Gene drive inoculation (additional to starting population size)
    if (nG>0) {
      g.df <- pop[rep(nrow(pop),nG),]
      rownames(g.df) <- g.df$id <- (pop$id[nrow(pop)]+1):(pop$id[nrow(pop)]+nG)
      g.df$aut.m <- -999 ## -999 for gene drive
    }
    pop <- rbind(pop,g.df)
    
    ## Autosome functionality for strategies 3 and 4 when positioned in exon (0 - nonfunctional, 1 - functional)
    if (strategy %in% 3:4 & position=='exon') {
      pop$aut.m.func <- ifelse(pop$aut.m<0,0,ifelse(runif(nrow(pop))>(1-pRnonFunc)^(nGuides-pop$aut.m),0,1))
      pop$aut.p.func <- ifelse(pop$aut.p<0,0,ifelse(runif(nrow(pop))>(1-pRnonFunc)^(nGuides-pop$aut.p),0,1))
      pop$aut.m <- ifelse(pop$aut.m%in%-(1:(nGuides+1)),-pop$aut.m-1,pop$aut.m) ## correct back to number of cutting sites
      pop$aut.p <- ifelse(pop$aut.p%in%-(1:(nGuides+1)),-pop$aut.p-1,pop$aut.p) ## correct back to number of cutting sites
    }
    
    ## correction for resistant sites through polymorphism
    pop$aut.m <- ifelse(pop$aut.m==-999,-999,rbinom(nrow(pop),pop$aut.m,1-Pr))
    pop$aut.p <- ifelse(pop$aut.p==-999,-999,rbinom(nrow(pop),pop$aut.p,1-Pr))
    
    ## Classification of phenotypic sex and reproductive individuals
    pop <- reproductive(pop, strategy=strategy)
    
    ## return
    pop
  }
  
  ## Reproduction (breeding, inheritance)
  reproduce <- function(pop, m){
    
    ## select breeders
    rep.f.ind <- which(pop$p.sex=='f' & pop$reprod==1) ## females
    rep.m.ind <- which(pop$p.sex=='m' & pop$reprod==1) ## males
    
    ## pairing
    if (length(rep.f.ind)>0 & length(rep.m.ind)>0) {
      
      rep.m.ind <- sample(rep.m.ind,length(rep.f.ind),replace=T) ## update males that breed by sampling with replacement
      
      ## final numbers of females and males mating
      (n.rep.f <- length(rep.f.ind))
      (n.rep.m <- length(rep.m.ind))
      n.offspr.vec <- rpois(n.rep.f, m) # how many offspring per reproducing female
      (n.offspr <- sum(n.offspr.vec))
      
    } else  {
      
      n.offspr <- 0
    }
    
    ## births
    if(n.offspr > 0){
      
      rep.f.df <- offspring <- pop[rep(rep.f.ind,n.offspr.vec),]
      rep.m.df <- pop[rep(rep.m.ind,n.offspr.vec),]
      f.aut.m.start <- rep.f.df$aut.m
      f.aut.p.start <- rep.f.df$aut.p
      m.aut.m.start <- rep.m.df$aut.m
      m.aut.p.start <- rep.m.df$aut.p
      offspring$id <- max(pop$id)+seq_len(n.offspr)
      offspring$age <- 0
      
      ## Germline homing
      
      ## mothers
      homing.ind.m <- which(rep.f.df$aut.m>0 & rep.f.df$aut.p==-999)
      if(length(homing.ind.m)>0) {
        rmulti <- t(apply(transit[match(rep.f.df$aut.m[homing.ind.m], rownames(transit)),,drop=F], 1, function(x) rmultinom(n=1,size=1,prob=x)))
        rep.f.df$aut.m[homing.ind.m] <- as.numeric(colnames(transit)[apply(rmulti,1,function(x) which(x==1))])
      }
      homing.ind.p <- which(rep.f.df$aut.m==-999 & rep.f.df$aut.p>0)
      if(length(homing.ind.p)>0) {
        rmulti <- t(apply(transit[match(rep.f.df$aut.p[homing.ind.p], rownames(transit)),,drop=F], 1, function(x) rmultinom(n=1,size=1,prob=x)))
        rep.f.df$aut.p[homing.ind.p] <- as.numeric(colnames(transit)[apply(rmulti,1,function(x) which(x==1))])
      }
      
      ## fathers
      homing.ind.m <- which(rep.m.df$aut.m>0 & rep.m.df$aut.p==-999)
      if(length(homing.ind.m)>0) {
        rmulti <- t(apply(transit[match(rep.m.df$aut.m[homing.ind.m], rownames(transit)),,drop=F], 1, function(x) rmultinom(n=1,size=1,prob=x)))
        rep.m.df$aut.m[homing.ind.m] <- as.numeric(colnames(transit)[apply(rmulti,1,function(x) which(x==1))])
      }
      homing.ind.p <- which(rep.m.df$aut.m==-999 & rep.m.df$aut.p>0)
      if(length(homing.ind.p)>0) {
        rmulti <- t(apply(transit[match(rep.m.df$aut.p[homing.ind.p], rownames(transit)),,drop=F], 1, function(x) rmultinom(n=1,size=1,prob=x)))
        rep.m.df$aut.p[homing.ind.p] <- as.numeric(colnames(transit)[apply(rmulti,1,function(x) which(x==1))])
      }
      
      ## Autosome allocation
      aut.m.rand <- runif(n.offspr)
      aut.p.rand <- runif(n.offspr)
      all.m.rand <- runif(n.offspr)
      all.p.rand <- runif(n.offspr)
      offspring$aut.m <- ifelse(aut.m.rand<0.5, rep.f.df$aut.m, rep.f.df$aut.p)
      offspring$aut.p <- ifelse(aut.p.rand<0.5, rep.m.df$aut.m, rep.m.df$aut.p)  
      offspring$all.m <- ifelse(all.m.rand<0.5, rep.f.df$all.m, rep.f.df$all.p)
      offspring$all.p <- ifelse(all.p.rand<0.5, rep.m.df$all.m, rep.m.df$all.p)
      
      ## Transfer functionality of autosomes
      ## and recalculate functionality of autosomes for alleles that were functional in the parents
      if (strategy %in% 3:4 & position=='exon') {
        
        ## transfer
        offspring$aut.m.func <- ifelse(aut.m.rand<0.5, rep.f.df$aut.m.func, rep.f.df$aut.p.func)
        aut.m.start <- ifelse(aut.m.rand<0.5, f.aut.m.start, f.aut.p.start)
        offspring$aut.p.func <- ifelse(aut.p.rand<0.5, rep.m.df$aut.m.func, rep.m.df$aut.p.func) 
        aut.p.start <- ifelse(aut.p.rand<0.5, m.aut.m.start, m.aut.p.start)
        aut.m.diff <- aut.m.start-offspring$aut.m ## loss of cutting sites
        aut.p.diff <- aut.p.start-offspring$aut.p ## loss of cutting sites

        ## recalculate
        offspring$aut.m.func <- ifelse(offspring$aut.m<0 | offspring$aut.m.func==0,0,ifelse(runif(nrow(offspring))>(1-pRnonFunc)^aut.m.diff,0,1))
        offspring$aut.p.func <- ifelse(offspring$aut.p<0 | offspring$aut.p.func==0,0,ifelse(runif(nrow(offspring))>(1-pRnonFunc)^aut.p.diff,0,1))
      }
      
      ## convert back to true number of cutting sites
      offspring$aut.m <- ifelse(offspring$aut.m%in%-(1:(nGuides+1)),-offspring$aut.m-1,offspring$aut.m) ## correct back to number of cutting sites
      offspring$aut.p <- ifelse(offspring$aut.p%in%-(1:(nGuides+1)),-offspring$aut.p-1,offspring$aut.p) ## correct back to number of cutting sites
      
      ## genetic sex allocation
      offspring$g.sex <- ifelse(offspring$all.p%in%'Y','m','f')
      
      ## if strategy 3, kill homozygous gene drive
      if (strategy==3) {
        if (position=='intron') {
          mort <- which(offspring$aut.m==-999 & offspring$aut.p==-999)
        } else {
          mort <- which(offspring$aut.m.func==0 & offspring$aut.p.func==0) ## if drive is in an exon, kill if no functional allele present
        }
        if(length(mort)>0) {
          offspring <- offspring[-mort,]  # kill these ones
        }
      }
      
      ## if strategy 3, and modelling compensation in utero, correct survival probability based on # implants
      if (strategy==3 & compensation) {
        offspring <- offspring[order(offspring$id),]
        implants <- as.numeric(table(offspring$id))
        implant.s <- implantSmax - (implantSmax-implantSmin)*((implants/m)^implantTheta)
        implant.s.vec <- rep(implant.s, implants)
        survivors <- rbinom(nrow(offspring),1,implant.s.vec)
        offspring <- offspring[survivors==1,]
      }
      
      ## fix ids and rownames
      offspring$id <- max(pop$id)+seq_len(nrow(offspring))
      rownames(offspring) <- offspring$id
      
      ## update reproduction status
      offspring <- reproductive(pop=offspring,strategy=strategy)
      
      # add offspring to population
      if (nrow(offspring)>0) {
        pop <- rbind(pop, offspring)
      }
    }
    pop
  }
  
  ## Mortality
  mortality <- function(pop, s){
    mort <- which(runif(nrow(pop)) > s) # kill these ones
    if (length(mort)>0) {
      pop <- pop[-mort,]
    }
    pop
  }
  
  # Ageing
  age <- function(pop){
    pop$age <- pop$age + genTime # advance age 
    pop
  }
  
  ## Supplementation
  supplementation <- function(pop, supplement) {
    supp.pop <- data.frame(id=tail(pop$id,1)+seq_len(supplement),age=2,g.sex='m',p.sex='m',
                           aut.m=-999,aut.p=rbinom(supplement,nGuides,1-Pr),all.m='X',all.p='Y',reprod=1)
    pop <- rbind(pop, supp.pop)
    pop
  }
  
  #################################################
  ## Create transition matrix for germline homing events
  #################################################
  
  GeneDriveSeqRec <- function(S, Pc, Pn) {
    
    mat <- matrix(0, S+1, S+1)
    rownames(mat) <- colnames(mat) <- 0:S
    diag(mat) <- sapply(0:S, function(v) (1-Pc)^v)
    n.vec <- unlist(apply(lower.tri(mat)[-1,,drop=F],2,function(x) which(x))) ## cutting sites remaining
    i.vec <- rep(0:(S-1), apply(lower.tri(mat)[-1,,drop=F],2,sum)[-(S+1)]) ## number of resulting sites
    mat[lower.tri(mat)] <- sapply(1:length(n.vec), function(x) choose(n.vec[x],i.vec[x])*((1-Pc)^i.vec[x])*(Pc^(n.vec[x]-i.vec[x]))*Pn^(n.vec[x]-i.vec[x]))
    
    ## add gene drive column
    mat <- cbind(mat, rep(0,S+1))
    colnames(mat)[ncol(mat)] <- -999
    n.vec2 <- 0:S ## cutting sites remaining
    mat[,'-999'] <- sapply(0:S,function(n) sum(sapply(1:n, function(i) choose(n,i)*(Pc^i)*((1-Pc)^(n-i))*(1-(Pn^i)))))
    
    return(mat)
  }
  GeneDriveSimRec <- function(S,Pc,Pn) {
    
    mat <- matrix(0,S,S+2)
    mat[1,1] <- Pc*Pn
    mat[1,2] <- (1-Pc)
    
    if (S>1) {
      for (nS in 2:S) {
        mat[nS,2:(nS+1)] <- (1-Pc)*mat[nS-1,1:nS]
        mat[nS,1:nS] <- mat[nS,1:nS] + Pc*Pn*mat[nS-1,1:nS]
        for (ni in 2:nS) {
          prob <- 0
          for (j in 2:ni) {
            prob <- prob + choose(ni-2,j-2)*Pc^j*(1-Pc)^(ni-j)*Pn
          }
          for (nl in seq(nS-ni,0,by=-1)) {
            if ((nS-ni)>0) {
              mat[nS,nS-ni-nl+1] <- mat[nS,nS-ni-nl+1] + prob*mat[nS-ni,nS-ni-nl+1]
            } else {
              mat[nS,1] = mat[nS,1] + prob
            }
          }
        }
      }
    }
    mat[,S+2] = 1 - rowSums(mat[,1:(S+1),drop=F])
    mat <- rbind(c(1,rep(0,S+1)), mat)
    rownames(mat) <- 0:S
    colnames(mat) <- c(0:S,'-999')
    return(mat)
  }
  
  ## create transition matrix depending on whether simulataneous/sequential cutting and position of gene drive
  ## when transType=='seq', intervening sequence deletion is indicated by -(s + 1)
  if (transType=='seq') {
    transit <- GeneDriveSeqRec(S=nGuides,Pc=Pc,Pn=Pn)
  } else if (transType=='sim' & position=='intron') {
    transit <- GeneDriveSimRec(S=nGuides,Pc=Pc,Pn=Pn)
  } else {
    transit.seq <- GeneDriveSeqRec(S=nGuides,Pc=Pc,Pn=Pn)
    transit.sim <- GeneDriveSimRec(S=nGuides,Pc=Pc,Pn=Pn)
    diff.mat <- (transit.sim-transit.seq)[,-(nGuides+2)]
    colnames(diff.mat) <- -(as.numeric(colnames(diff.mat))+1)
    transit <- cbind(diff.mat,transit.seq)
    transit[,'-999'] <- 1-rowSums(transit[,-ncol(transit)])
  }
  
  #############################################################################
  ### Derive survival for stable population and starting population vector
  #############################################################################
  
  ## calculate fertility rate for r=0
  r0.func <- function(s) {
    a <- matrix(0,4,4)
    a[1,1:2] <- s*m*0.5
    a[3,1:2] <- s*m*0.5
    a[2,1:2] <- a[4,3:4] <- s
    minim <- abs(r.func(a))
    return(minim)
  }
  (s <- optimise(r0.func,interval=c(1e-10,100))$minimum)
  
  ## get stable age distribution
  a <- matrix(0,4,4)
  a[1,1:2] <- s*m*0.5
  a[3,1:2] <- s*m*0.5
  a[2,1:2] <- a[4,3:4] <- s
  (sad <- stable.age.dist(a))
  
  ### adjust rmax for generation time
  (rmax.adj <- rmax*(genTime/52))
  
  ## calculate survival rate for r=rmax
  rmax.func <- function(s) {
    a <- matrix(0,4,4)
    a[1,1:2] <- s*m*0.5
    a[3,1:2] <- s*m*0.5
    a[2,1:2] <- a[4,3:4] <- s
    minim <- abs(r.func(a)-rmax.adj)
    return(minim)
  }
  (s.max <- optimise(rmax.func,interval=c(1e-10,100))$minimum)
  
  ## intercept and slope of logistic survival equation
  (alpha.s <- -log((1/s.max)-1)) 
  (beta.s <- -alpha.s - log((1/s)-1)) 
  
  ## starting n vector
  nInit <- K
  (nInit.vec <- round(sad*nInit))
  
  ###################
  ### Simulation ####
  ###################
  
  # Object for storing results (population size in time)
  year <- seq(0,nYear,by=genTime/52)
  gen <- 0:(nYear*(52/genTime))
  nT <- length(gen)-1
  var <- c('N','nF','nM','propG','propN')
  results <- array(NA, dim=c(length(year), length(var), nIter), dimnames=list(year=year,var=var,iter=1:nIter))
  results[1,,] <- c(nInit+nG, (nInit/2), (nInit/2)+nG, nG/(nInit+nG), 0)
  
  # Run 
  for (iter in 1:nIter) {
    
    # initial population
    N <- nInit+nG
    pop <- create(nInit=nInit, nInit.vec=nInit.vec, nG=nG, Pr=Pr)
    
    for(t in 2:(nT+1)) {
      
      # supplementation each year (except first year)
      if (supplement>0 & t>2) {
        decTime <- year[t-2]
        if (decTime==ceiling(decTime)) {
          pop <- supplementation(pop=pop, supplement=supplement)
        }
      }
      
      # demography
      pop <- reproduce(pop, m=m)
      (N.now <- results[match(floor(gen[t-1]), gen),'N',iter])
      (s.now <- ifelse(is.na(rmax),s,1/(1+exp(-(alpha.s+beta.s*N.now/K)))))
      pop <- mortality(pop, s=s.now)
      pop <- age(pop)

      ## calculate summaries
      (N <- nrow(pop))
      nF <- sum(pop$p.sex=='f')
      nM <- sum(pop$p.sex=='m')
      propG <- sum(pop$aut.m==-999 | pop$aut.p==-999)/N ## at least one gene drive allele
      propN <- sum(pop$aut.m==0 | pop$aut.p==0)/N ## at least resistance allele
      
      # store results (just for each whole generation time)
      results[t,'N',iter] <- N
      results[t,'nF',iter] <- nF
      results[t,'nM',iter] <- nM
      results[t,'propG',iter] <- propG
      results[t,'propN',iter] <- propN
      
      # break when one sex remains or N = 0
      if(nF==0 | nM==0 | N==0 ) {break}
      
    }
    
    print(iter)
  }
  
  ## melt into required format
  N.res <- melt(results[,'N',],varnames=c('year','iter'),value.name='N')
  nF.res <- melt(results[,'nF',],varnames=c('year','iter'),value.name='nF')
  nM.res <- melt(results[,'nM',],varnames=c('year','iter'),value.name='nM')
  propG.res <- melt(results[,'propG',],varnames=c('year','iter'),value.name='propG')
  propN.res <- melt(results[,'propN',],varnames=c('year','iter'),value.name='propN')
  res <- cbind(N.res,nF.res[,'nF',drop=F],nM.res[,'nM',drop=F],propG.res[,'propG',drop=F],propN.res[,'propN',drop=F])
  
  ## add all the parameters to the results data frame too
  for (i in 1:ncol(input)) {
    res[[names(input)[i]]] <- input[,i]
  }
  return(res)
}


############################################
## compile the function so it runs faster
############################################

gd.comp <- cmpfun(gd)

############################################################################
## create data frame of parameter inputs
############################################################################

input <- list()
input$strategy <- 1:4 ## 1 = Heterozygotic XX sterility
                      ## 2 = Heterozygotic XX sex reversal
                      ## 3 = Homozygotic embryonic non-viability
                      ## 4 = Homozygotic XX sterility
input$nYear <- 10 # maximum number of years
input$nIter <- 10 # number of iterations
input$K <- 2500 # carrying capacity of offspring
input$nG <- 5 # initial number carrying the gene drive
input$supplement <- 0 # 0 (no supplementation) or 1 (annual supplementation)
input$nGuides <- c(1,3,5) # number of guide RNAs
input$Pc <- 0.95 # cutting probability
input$Pn <- 0.02 # NHEJ probability
input$m <- 6 # litter size
input$genTime <- 5.2 # generation time (in weeks)
input$rmax <- 7.76 # maximum annual population growth rate
input$Pr <- 0 # probability of inherent resistance (affects number of susceptible sites at initiation)
input$pRnonFunc <- 2/3 # probability of single-site deletion causing loss-of-function (if position=='exon')
input$compensation <- FALSE ## reproductive compensation for strategy 3 only
input$implantSmax <- 1 # maximum embryonic survival rate (when compensation==TRUE)
input$implantSmin <- 1 # minimum embryonic survival rate (when compensation==TRUE)
input$implantTheta <- 1 # theta parameter for theta-logistic compensation (when compensation==TRUE)
input$position <- c('intron','exon') # position of gene drive, one of 'intron' or 'exon'
input$transType <- c('seq','sim') ## transition matrix type, either 'seq' (sequential) or 'sim' (simultaneous)

inputs <- expand.grid(input)
inputs <- subset(inputs, !(nGuides>1 & Pn==0))
inputs <- subset(inputs, !((strategy%in%1:2 & position=='exon') | (strategy%in%3:4 & position=='intron')))
inputs <- subset(inputs, !(implantSmax==0.6 & implantTheta==3))
inputs$supplement <- ifelse(inputs$supplement==1,inputs$nG,inputs$supplement)
inputs$species <- 'mouse'
inputs <- inputs[order(inputs$strategy, inputs$nGuides),]
rownames(inputs) <- 1:nrow(inputs)
inputs

#########################################################
## run models
## and plot % extinct populations over time
#########################################################

## run in parallel
res <- foreach(rowNum=1:nrow(inputs), .packages='reshape2',.verbose=F) %dopar% {gd.comp(input=inputs[rowNum,])}

## collate and calculate % extinct populations over time
res <- do.call(rbind, res)
res$N[is.na(res$N)] <- 0
percExtinct <- aggregate(N ~ year + strategy + nGuides + transType, data=res, FUN=function(x) 100*sum(x==0)/length(x))
names(percExtinct)[ncol(percExtinct)] <- 'PE'

## plot
percExtinct$nGuides <- factor(percExtinct$nGuides)
ggplot(percExtinct, aes(year, PE, group=nGuides, colour=nGuides)) +
  facet_grid(transType~strategy) + geom_line(size=1.2) + xlab('Year') + ylab('% Eradicated populations') +
  theme(title=element_text(size=20), axis.text=element_text(size=15),axis.title=element_text(size=20),legend.text=element_text(size=15),
        strip.text.x = element_text(size = 15), strip.text.y = element_text(size = 15))

